# Copyright (C) 2025 ETH Zurich (Mohit Pundir)
#
# This file is part of tatva.
#
# tatva is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tatva is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with tatva.  If not, see <https://www.gnu.org/licenses/>.


from __future__ import annotations

from collections.abc import Sequence
from functools import partial
from typing import Callable, Generic, ParamSpec, Protocol, TypeAlias, TypeVar, cast

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
from jax_autovmap import autovmap

from tatva.element import Element
from tatva.mesh import Mesh, find_containing_polygons

P = ParamSpec("P")
RT = TypeVar("RT", bound=jax.Array | tuple, covariant=True)
ElementT = TypeVar("ElementT", bound=Element)
Numeric: TypeAlias = float | int | jnp.number


class MappableOverElementsAndQuads(Protocol[P, RT]):
    """Internal protocol for functions that are mapped over elements using
    `Operator.map`."""

    @staticmethod
    def __call__(
        xi: jax.Array,
        *el_values: P.args,
        **el_kwargs: P.kwargs,
    ) -> RT: ...


MappableOverElements: TypeAlias = Callable[P, RT]


class MappedCallable(Protocol[P, RT]):
    @staticmethod
    def __call__(
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> RT: ...


class Operator(Generic[ElementT], eqx.Module):
    """A class that provides an Operator for finite element method (FEM) assembly.

    Args:
        mesh: The mesh containing the elements and nodes.
        element: The element type used for the finite element method.

    Provides several operators for evaluating and integrating functions over the mesh,
    such as `integrate`, `eval`, and `grad`. These operators can be used to compute
    integrals, evaluate functions at quadrature points, and compute gradients of
    functions at quadrature points.

    Example:
        >>> from tatva import Mesh, Tri3, Operator
        >>> mesh = Mesh.unit_square(10, 10)  # Create a mesh
        >>> element = Tri3()  # Define an element type
        >>> operator = Operator(mesh, element)
        >>> nodal_values = jnp.array(...)  # Nodal values at the mesh nodes
        >>> energy = operator.integrate(energy_density)(nodal_values)
    """

    mesh: Mesh
    element: ElementT

    def __post_init__(self) -> None:
        """Validates the mesh and element compatibility. Does a series of checks to ensure
        that the mesh and element are useable together.

        Raises:
            ValueError: If the mesh or element are not compatible.
            TypeError: If the mesh element connectivity is not of integer type.
        """
        coords = self.mesh.coords
        elements = self.mesh.elements
        quad_points = self.element.quad_points

        if coords.ndim != 2:
            raise ValueError(
                "Mesh coordinates must be a 2D array shaped (n_nodes, n_dim)."
            )
        if coords.shape[0] == 0:
            raise ValueError("Mesh must contain at least one node.")

        if elements.ndim != 2:
            raise ValueError(
                "Mesh elements must be a 2D array shaped (n_elements, n_nodes_per_element)."
            )
        if elements.shape[0] == 0:
            raise ValueError("Mesh must contain at least one element.")
        if not jnp.issubdtype(elements.dtype, jnp.integer):
            raise TypeError("Mesh element connectivity must contain integer indices.")

        if quad_points.ndim != 2 or quad_points.shape[0] == 0:
            raise ValueError(
                "Element must define at least one quadrature point in an (n_q, n_dim) array."
            )

        local_dim = quad_points.shape[1]
        global_dim = coords.shape[1]
        if local_dim > 1 and local_dim != global_dim:
            raise ValueError(
                f"Element {self.element.__class__.__name__} expects {local_dim}D coordinates but mesh provides {global_dim}D nodes."
            )
        if local_dim == 0:
            raise ValueError("Element must have a positive number of local dimensions.")

        n_nodes_per_element = elements.shape[1]
        shape_fn = np.asarray(self.element.shape_function(self.element.quad_points[0]))
        if shape_fn.ndim != 1:
            raise ValueError(
                "Element shape function must return a 1D array of nodal weights."
            )
        if shape_fn.shape[0] != n_nodes_per_element:
            raise ValueError(
                f"Mesh connectivity lists {n_nodes_per_element} nodes per element but {self.element.__class__.__name__} expects {shape_fn.shape[0]}."
            )

        flat_elements = elements.ravel()
        if flat_elements.min() < 0:
            raise ValueError(
                "Mesh element connectivity contains negative node indices."
            )
        if flat_elements.max() >= coords.shape[0]:
            raise ValueError(
                "Mesh element connectivity references nodes outside the mesh coordinates array."
            )

    def _vmap_over_elements_and_quads(
        self, nodal_values: jax.Array, func: MappableOverElementsAndQuads
    ) -> jax.Array:
        """Helper function. Maps a function over the elements and quadrature points of the
        mesh.

        Args:
            nodal_values: The nodal values at the element's nodes (shape: (n_nodes, n_values))
            func: The function to map over the elements and quadrature points.

        Returns:
            A jax.Array with the results of the function applied at each quadrature point
            of each element (shape: (n_elements, n_quad_points, n_values)).
        """

        def _at_each_element(
            el_nodal_values: jax.Array, el_nodal_coords: jax.Array
        ) -> jax.Array:
            return eqx.filter_vmap(
                partial(
                    func,
                    el_nodal_values=el_nodal_values,
                    el_nodal_coords=el_nodal_coords,
                )
            )(self.element.quad_points)

        return eqx.filter_vmap(
            _at_each_element,
            in_axes=(0, 0),
        )(nodal_values[self.mesh.elements], self.mesh.coords[self.mesh.elements])

    def map(
        self,
        func: MappableOverElementsAndQuads[P, RT],
        *,
        element_quantity: Sequence[int] = (),
    ) -> MappedCallable[P, RT]:
        """Maps a function over the elements and quad points of the mesh.

        Returns a function that takes values at nodal points (globally) and returns the
        vmapped result over the elements and quad points.

        Args:
            func: The function to map over the elements and quadrature points.
            element_quantity: Indices of the arguments of `func` that are quantities
                defined per element. The rest of the arguments are assumed to be defined
                at nodal points.
        """

        def _mapped(*values: P.args, **kwargs: P.kwargs) -> RT:
            # values should be arrays!
            _values = cast(tuple[jax.Array, ...], values)

            def _at_each_element(*el_values) -> jax.Array:
                return eqx.filter_vmap(
                    lambda xi: func(xi, *el_values, **kwargs),
                )(self.element.quad_points)

            return eqx.filter_vmap(
                _at_each_element,
                in_axes=(0,) * len(values),
            )(
                *(
                    v[self.mesh.elements] if i not in element_quantity else v
                    for i, v in enumerate(_values)
                )
            )

        return _mapped

    def map_over_elements(
        self,
        func: MappableOverElements[P, RT],
        *,
        element_quantity: Sequence[int] = (),
    ) -> MappedCallable[P, RT]:
        """Maps a function over the elements of the mesh.

        Returns a function that takes values at nodal points (globally) and returns the
        vmapped result over the elements.

        Args:
            func: The function to map over the elements.
            element_quantity: Indices of the arguments of `func` that are quantities
                defined per element. The rest of the arguments are assumed to be defined
                at nodal points.
        """

        def _mapped(*values: P.args, **kwargs: P.kwargs) -> RT:
            # values should be arrays!
            _values = cast(tuple[jax.Array, ...], values)

            def _at_each_element(*el_values) -> RT:
                return func(*el_values, **kwargs)

            return eqx.filter_vmap(
                _at_each_element,
                in_axes=(0,) * len(values),
            )(
                *(
                    v[self.mesh.elements] if i not in element_quantity else v
                    for i, v in enumerate(_values)
                )
            )

        return _mapped

    def integrate(self, arg: jax.Array | Numeric) -> jax.Array:
        """Integrate a nodal_array, quad_array, or numeric value over the mesh.

        Args:
            arg: An array of nodal values (shape: (n_nodes, n_values)), an array of
                quadrature values (shape: (n_elements, n_quad_points, n_values)), or a
                numeric value (float or int).

        Returns:
            The integral of the nodal values or quadrature values over the mesh.
        """
        res = self.integrate_per_element(arg)
        return jnp.sum(res, axis=(0,))  # Sum over elements and quadrature points

    def integrate_per_element(self, arg: jax.Array | Numeric) -> jax.Array:
        """Integrate a nodal_array, quad_array, or numeric value over the mesh. Returning the
        integral per element.

        Args:
            arg: An array of nodal values (shape: (n_nodes, n_values)), an array of
                quadrature values (shape: (n_elements, n_quad_points, n_values)), or a
                numeric value (float or int).

        Returns:
            A `jax.Array` where each element contains the integral of the values in the
            element (shape: (n_elements, n_values)).
        """
        if isinstance(arg, Numeric):
            res = self._integrate_quad_array(self.eval(jnp.array([arg])))
        elif arg.shape[0] == self.mesh.elements.shape[0]:  # element field
            res = self._integrate_quad_array(arg)
        else:  # nodal field
            field_at_quads = self.eval(arg)
            res = self._integrate_quad_array(field_at_quads)

        return res

    def _integrate_quad_array(self, quad_values: jax.Array) -> jax.Array:
        """Integrates a given array of values at quadrature points over the mesh.

        Args:
            quad_values: The values at the quadrature points
                (shape: (n_elements, n_quad_points, n_values))

        Returns:
            A `jax.Array` where each element contains the integral of the values in the
            element (shape: (n_elements, n_values)).
        """

        def _get_det_J(xi: jax.Array, el_nodal_coords: jax.Array) -> jax.Array:
            """Calls the function element.get_jacobian and returns the second output."""
            return self.element.get_jacobian(xi, el_nodal_coords)[1]

        det_J_elements = self.map(_get_det_J)(self.mesh.coords)

        return jnp.einsum(
            "eq...,eq->e...",
            quad_values,
            jnp.einsum("eq,q->eq", det_J_elements, self.element.quad_weights),
        )

    def eval(self, nodal_values: jax.Array) -> jax.Array:
        """Evaluates the nodal values at the quadrature points.

        Args:
            nodal_values: The nodal values at the element's nodes (shape: (n_nodes, n_values))

        Returns:
            A `jax.Array` with the values of the nodal values at each quadrature point of
            each element (shape: (n_elements, n_quad_points, n_values)).
        """

        def _eval_quad(
            xi: jax.Array, el_nodal_values: jax.Array, el_nodal_coords: jax.Array
        ) -> jax.Array:
            """Calls the function (interpolator) on a quad point."""
            return self.element.interpolate(xi, el_nodal_values)

        return self._vmap_over_elements_and_quads(nodal_values, _eval_quad)

    def grad(self, nodal_values: jax.Array) -> jax.Array:
        """Computes the gradient of the nodal values at the quad points.

        Args:
            nodal_values: The nodal values at the element's nodes (shape: (n_nodes, n_values))

        Returns:
            A `jax.Array` with the gradient of the nodal values at each quadrature point
            of each element (shape: (n_elements, n_quad_points, n_values, n_dim)).
        """

        def _gradient_quad(
            xi: jax.Array, el_nodal_values: jax.Array, el_nodal_coords: jax.Array
        ) -> jax.Array:
            """Calls the function (gradient) on a quad point."""
            u_grad = self.element.gradient(xi, el_nodal_values, el_nodal_coords)
            return u_grad

        return self._vmap_over_elements_and_quads(nodal_values, _gradient_quad)

    def interpolate(self, arg: jax.Array, points: jax.Array) -> jax.Array:
        """Interpolates nodal values to a set of points in the physical space.

        Args:
            arg: The nodal values to interpolate.
            points: The points to interpolate the function or nodal values to.

        Returns:
            A `jax.Array` with the interpolated values at the given points.
        """

        @jax.jit
        def compute_rhs(point: jax.Array, nodal_coords: jax.Array) -> jax.Array:
            xi0 = self.element.quad_points[0]
            x0, _, _ = self.element.get_local_values(xi0, nodal_coords, nodal_coords)
            return x0 - point

        @jax.jit
        def compute_lhs(nodal_coords: jax.Array) -> jax.Array:
            dfdxi = jax.jacrev(self.element.get_local_values)
            return dfdxi(self.element.quad_points[0], nodal_coords, nodal_coords)[0]

        @autovmap(point=1, nodal_coords=2)
        def _map_physical_to_reference(
            point: jax.Array, nodal_coords: jax.Array
        ) -> jax.Array:
            rhs = compute_rhs(point, nodal_coords)
            lhs = compute_lhs(nodal_coords)
            delta_xi = jnp.linalg.solve(lhs, -rhs)
            return self.element.quad_points[0] + delta_xi

        def map_physical_to_reference(points: jax.Array) -> tuple[jax.Array, jax.Array]:
            element_indices = find_containing_polygons(
                points, self.mesh.coords[self.mesh.elements]
            )
            valid_indices = element_indices != -1
            return (
                _map_physical_to_reference(
                    points[valid_indices],
                    self.mesh.coords[
                        self.mesh.elements[element_indices[valid_indices]]
                    ],
                ),
                self.mesh.elements[element_indices[valid_indices]],
            )

        valid_quad_points, valid_elements = map_physical_to_reference(points)

        if valid_quad_points.shape[0] != points.shape[0]:
            raise RuntimeError("Some points are outside the mesh, revise the points")

        return self._interpolate_direct(arg, valid_quad_points, valid_elements)

    def _interpolate_direct(
        self,
        nodal_values: jax.Array,
        valid_quad_points: jax.Array,
        valid_elements: jax.Array,
    ) -> jax.Array:
        """Interpolates the given nodal values at the quad points.

        Args:
            nodal_values: The nodal values at the element's nodes (shape: (n_nodes, n_values))
            valid_quad_points: The quadrature points in the reference element
                (shape: (n_valid_points, n_dim))
            valid_elements: The indices of the elements containing the quadrature points
                (shape: (n_valid_points,))

        Returns:
            A `jax.Array` with the values of the nodal values at each quadrature point of
            each element (shape: (n_valid_points, n_values)).
        """

        def _interpolate_quad(
            xi: jax.Array, el_nodal_values: jax.Array, el_nodal_coords: jax.Array
        ) -> jax.Array:
            """Calls the function (interpolator) on a quad point."""
            return self.element.interpolate(xi, el_nodal_values)

        return eqx.filter_vmap(
            _interpolate_quad,
            in_axes=(0, 0, 0),
        )(
            valid_quad_points,
            nodal_values[valid_elements],
            self.mesh.coords[valid_elements],
        )



import jax
import numpy as np
import pytest

jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp

from tatva.compound import Compound, field


class SimpleState(Compound):
    u = field((2, 3))
    phi = field((2,))


class StackedState(Compound, stack_fields=("u", "v"), stack_axis=-1):
    u = field((2, 2))
    v = field((2, 2))
    w = field((2,))


@pytest.mark.parametrize("state_cls", [SimpleState, StackedState])
def test_compound_size_matches_flat_array(state_cls):
    state = state_cls()
    assert state.arr.shape == (state_cls.size,)
    assert jnp.all(state.arr == 0)


def test_compound_field_access_and_assignment():
    state = SimpleState()
    assert SimpleState.size == 8

    np.testing.assert_array_equal(state.u, jnp.zeros((2, 3)))
    np.testing.assert_array_equal(state.phi, jnp.zeros((2, 1)))

    u_val = jnp.arange(6.0).reshape(2, 3)
    phi_val = jnp.array([10.0, 20.0])

    state.u = u_val
    state.phi = phi_val

    np.testing.assert_array_equal(state.u, u_val)
    np.testing.assert_array_equal(state.phi, phi_val.reshape(2, 1))
    np.testing.assert_array_equal(state.arr[:6], u_val.flatten())
    np.testing.assert_array_equal(state.arr[6:], phi_val.flatten())

    assert len(state) == 2
    shapes = [component.shape for component in state]
    assert shapes == [(2, 3), (2, 1)]
    assert repr(state) == "SimpleState(u=(2, 3), phi=(2, 1))"


def test_compound_index_helpers():
    np.testing.assert_array_equal(np.array(SimpleState.u[1]), np.array([3, 4, 5]))
    np.testing.assert_array_equal(np.array(SimpleState.u[:, 1]), np.array([1, 4]))
    np.testing.assert_array_equal(np.array(SimpleState.phi[1]), np.array([7]))

    np.testing.assert_array_equal(np.array(SimpleState[0]), np.array([0, 1, 2, 6]))
    np.testing.assert_array_equal(np.array(SimpleState[1]), np.array([3, 4, 5, 7]))
    np.testing.assert_array_equal(np.array(SimpleState[1, 2]), np.array([5]))


def test_pytree_roundtrip_and_addition():
    arr_a = jnp.arange(SimpleState.size, dtype=jnp.float64)
    arr_b = jnp.arange(SimpleState.size, dtype=jnp.float64) * 2

    state_a = SimpleState(arr_a)
    state_b = SimpleState(arr_b)

    leaves, tree_def = jax.tree_util.tree_flatten(state_a)
    assert len(leaves) == 1
    np.testing.assert_array_equal(leaves[0], arr_a)

    rebuilt = jax.tree_util.tree_unflatten(tree_def, leaves)
    assert isinstance(rebuilt, SimpleState)
    np.testing.assert_array_equal(rebuilt.arr, arr_a)
    np.testing.assert_array_equal(rebuilt.u, state_a.u)
    np.testing.assert_array_equal(rebuilt.phi, state_a.phi)

    summed = state_a + state_b
    np.testing.assert_array_equal(summed.arr, arr_a + arr_b)
    np.testing.assert_array_equal(summed.u, state_a.u + state_b.u)
    np.testing.assert_array_equal(summed.phi, state_a.phi + state_b.phi)


def test_stack_fields_access_and_indices():
    state = StackedState(jnp.arange(StackedState.size, dtype=jnp.float64))

    expected_u = jnp.array([[0.0, 1.0], [4.0, 5.0]])
    expected_v = jnp.array([[2.0, 3.0], [6.0, 7.0]])
    expected_w = jnp.array([[8.0], [9.0]])

    np.testing.assert_array_equal(state.u, expected_u)
    np.testing.assert_array_equal(state.v, expected_v)
    np.testing.assert_array_equal(state.w, expected_w)

    np.testing.assert_array_equal(np.array(StackedState.u[1]), np.array([4, 5]))
    np.testing.assert_array_equal(np.array(StackedState.v[0]), np.array([2, 3]))

    np.testing.assert_array_equal(np.array(StackedState[0]), np.array([0, 1, 2, 3, 8]))
    np.testing.assert_array_equal(np.array(StackedState[1]), np.array([4, 5, 6, 7, 9]))



# Copyright (C) 2025 ETH Zurich (Mohit Pundir)
#
# This file is part of tatva.
#
# tatva is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tatva is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with tatva.  If not, see <https://www.gnu.org/licenses/>.


# pyright: reportUnusedImport=false

from .base import Element as Element
from .base import Hexahedron8, Line2, Quad4, Tetrahedron4, Tri3



# Copyright (C) 2025 ETH Zurich (Mohit Pundir)
#
# This file is part of tatva.
#
# tatva is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tatva is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with tatva.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Literal, Optional

import numpy as np
from jax import Array

try:
    import matplotlib.pyplot as plt
    from matplotlib.collections import PolyCollection
    from matplotlib.tri import Triangulation
    from mpl_toolkits.axes_grid1 import make_axes_locatable
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "tatva.plotting requires the optional dependency 'matplotlib'. "
        "Install it with `pip install tatva[plotting]` or `pip install matplotlib`."
    ) from exc

if TYPE_CHECKING:
    from matplotlib.axes import Axes
else:
    Axes = object


try:
    from ._pyvista import get_pyvista_grid as get_pyvista_grid
except ImportError:
    get_pyvista_grid = None


def plot_element_values(
    coords: Array,
    elements: Array,
    values: Array,
    *,
    u: Optional[Array] = None,
    ax: Optional[Axes] = None,
    scale: float = 1.0,
    label: Optional[str] = None,
    cmap="managua_r",
    **kwargs,
):
    """
    Plot the element values of a field on a mesh.

    Args:
        coords: The coordinates of the mesh nodes.
        elements: The connectivity of the mesh elements.
        values: The element values to plot.
        u: The displacement field.
        ax: The axes to plot on.
        scale: The scale of the displacement field.
        label: The label of the colorbar.
        cmap: The colormap to use.
        **kwargs : Additional keyword arguments to pass to PolyCollection.
    """

    if u is not None:
        coords = coords + scale * u

    vertices = coords[elements]

    if ax is None:
        fig = plt.figure(figsize=(5, 4))
        ax = fig.add_subplot(111)

    poly = PolyCollection(vertices, cmap=cmap, linewidths=0.3, **kwargs)  # pyright: ignore[reportArgumentType]
    poly.set_array(values.flatten())
    poly.set_edgecolor("face")
    ax.add_collection(poly)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("top", size="10%", pad=0.2)
    plt.colorbar(poly, cax=cax, label=label, orientation="horizontal", location="top")


def plot_nodal_values(
    coords: Array,
    elements: Array,
    values: Array,
    *,
    u: Optional[Array] = None,
    ax: Optional[Axes] = None,
    scale: float = 1.0,
    label: Optional[str] = None,
    cmap="managua_r",
    edgecolors: str = "none",
    shading: Literal["flat", "gouraud"] = "gouraud",
    **kwargs,
):
    """
    Plot the nodal values of a field on a mesh.

    Args:
        coords: The coordinates of the mesh nodes.
        elements: The connectivity of the mesh elements.
        values: The element values to plot.
        u: The displacement field.
        ax: The axes to plot on.
        scale: The scale of the displacement field.
        label: The label of the colorbar.
        cmap: The colormap to use.
        edgecolors: The edge colors of the triangles.
        shading: The shading of the triangles.
        **kwargs: Additional keyword arguments to pass to tripcolor.
    """

    coords_np = np.asarray(coords)
    if u is not None:
        coords_np = coords_np + scale * np.asarray(u)

    if ax is None:
        fig = plt.figure(figsize=(5, 4))
        ax = fig.add_subplot(111)

    elements_np = np.asarray(elements)

    if elements_np.shape[1] == 3:
        triangles = elements_np
    elif elements_np.shape[1] == 4:
        # Split each quad into two triangles along the (0, 2) diagonal.
        first_half = elements_np[:, [0, 1, 2]]
        second_half = elements_np[:, [0, 2, 3]]
        triangles = np.concatenate((first_half, second_half), axis=0)
    else:
        raise ValueError(
            "Only triangular or quadrilateral elements are supported for nodal values."
        )

    triangles = np.asarray(triangles, dtype=int)
    tri = Triangulation(coords_np[:, 0], coords_np[:, 1], triangles)

    cb = ax.tripcolor(
        tri,
        np.asarray(values),
        shading=shading,
        cmap=cmap,
        edgecolors=edgecolors,
        **kwargs,
    )

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("top", size="10%", pad=0.2)
    plt.colorbar(cb, cax=cax, label=label, orientation="horizontal", location="top")



# Copyright (C) 2025 ETH Zurich (Mohit Pundir)
#
# This file is part of tatva.
#
# tatva is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tatva is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with tatva.  If not, see <https://www.gnu.org/licenses/>.


from typing import Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as np
import sparsejac
from jax import Array
from jax.experimental import sparse as jax_sparse

from tatva import Mesh


def jacfwd(func, sparsity_pattern):
    return sparsejac.jacfwd(func, sparsity=sparsity_pattern)


def jacrev(func, sparsity_pattern):
    return sparsejac.jacrev(func, sparsity=sparsity_pattern)


def _create_sparse_structure(elements, n_dofs_per_node, K_shape):
    """
    Create a sparse structure for a given set of elements and constraints.
    Args:
        elements: (num_elements, nodes_per_element)
        n_dofs_per_node: Number of degrees of freedom per node
        K_shape: Shape of the matrix K
    Returns:
        data: (num_nonzero_elements,)
        indices: (num_nonzero_elements, 2)
    """

    elements = jnp.array(elements)
    num_elements, nodes_per_element = elements.shape

    # Compute all (i, j, k, l) combinations for each element
    i_idx = jnp.repeat(
        elements, nodes_per_element, axis=1
    )  # (num_elements, nodes_per_element^2)
    j_idx = jnp.tile(
        elements, (1, nodes_per_element)
    )  # (num_elements, nodes_per_element^2)

    # Expand for n_dofs_per_node
    k_idx = jnp.arange(n_dofs_per_node, dtype=jnp.int32)
    l_idx = jnp.arange(n_dofs_per_node, dtype=jnp.int32)
    k_idx, l_idx = jnp.meshgrid(k_idx, l_idx, indexing="ij")
    k_idx = k_idx.flatten()
    l_idx = l_idx.flatten()

    # For each element, get all (row, col) indices
    def element_indices(i, j):
        row = n_dofs_per_node * i + k_idx
        col = n_dofs_per_node * j + l_idx
        return row, col

    # Vectorize over all (i, j) pairs for all elements
    row_idx, col_idx = jax.vmap(element_indices)(i_idx.flatten(), j_idx.flatten())

    # Flatten and clip to matrix size
    row_idx = row_idx.flatten()
    col_idx = col_idx.flatten()
    mask = (row_idx < K_shape[0]) & (col_idx < K_shape[1])
    row_idx = row_idx[mask]
    col_idx = col_idx[mask]

    # Create the sparse structure
    indices = np.unique(np.vstack((row_idx, col_idx)).T, axis=0)

    data = np.ones(indices.shape[0], dtype=jnp.int32)
    sparsity_pattern = jax_sparse.BCOO((data, indices.astype(np.int32)), shape=K_shape)  # type: ignore
    return sparsity_pattern


def get_bc_indices(sparsity_pattern: jax_sparse.BCOO, fixed_dofs: Array):
    """
    Get the indices of the fixed degrees of freedom.
    Args:
        sparsity_pattern: jax.experimental.sparse.BCOO
        fixed_dofs: (num_fixed_dofs,)
    Returns:
        zero_indices: (num_zero_indices,)
        one_indices: (num_one_indices,)
    """

    indices = sparsity_pattern.indices
    zero_indices = []
    one_indices = []

    for dof in fixed_dofs:
        indexes = np.where(indices[:, 0] == dof)[0]
        for idx in indexes:
            zero_indices.append(int(idx))

        idx = np.where(np.all(indices == np.array([dof, dof]), axis=1))[0][0]
        one_indices.append(int(idx))

    return np.array(zero_indices), np.array(one_indices)


def create_sparsity_pattern(
    mesh: Mesh,
    n_dofs_per_node: int,
    K_shape: Optional[Tuple[int, int]] = None,
    constraint_elements: Optional[Array] = None,
):
    """
    Create a sparsity pattern for a given set of elements and constraints.
    Args:
        mesh: Mesh object
        n_dofs_per_node: Number of degrees of freedom per node
        constraint_elements: Optional array of constraint elements. If provided, the sparsity pattern will be created for the constraint elements.
    Returns:
        sparsity_pattern: jax.experimental.sparse.BCOO
    """

    elements = mesh.elements

    if K_shape is None:
        K_shape = (
            n_dofs_per_node * mesh.coords.shape[0],
            n_dofs_per_node * mesh.coords.shape[0],
        )

    sparsity_pattern = _create_sparse_structure(elements, n_dofs_per_node, K_shape)
    if constraint_elements is not None:
        sparsity_pattern_constraint = _create_sparse_structure(
            constraint_elements, n_dofs_per_node, K_shape
        )

        combined_data = np.concatenate(
            [sparsity_pattern.data, sparsity_pattern_constraint.data]
        )
        combined_indices = np.concatenate(
            [sparsity_pattern.indices, sparsity_pattern_constraint.indices]
        )
        sparsity_pattern = jax_sparse.BCOO(
            (combined_data, combined_indices),  # type: ignore
            shape=K_shape,
        )

    return sparsity_pattern


def create_sparsity_pattern_KKT(mesh: Mesh, n_dofs_per_node: int, B: Array):
    """
    Create a sparsity pattern for the KKT system.
    Args:
        mesh: Mesh object
        n_dofs_per_node: Number of degrees of freedom per node
        B: Constraint matrix (nb_cons, n_dofs)
    Returns:
        sparsity_pattern_KKT: jax.experimental.sparse.BCOO
    """

    nb_cons = B.shape[0]

    K_sparsity_pattern = create_sparsity_pattern(mesh, n_dofs_per_node=n_dofs_per_node)
    B_sparsity_pattern = jax_sparse.BCOO.fromdense(B).astype(jnp.int32)

    sparsity_pattern_left = jax_sparse.bcoo_concatenate(
        [K_sparsity_pattern, B_sparsity_pattern], dimension=0
    )

    BT_sparsity_pattern = jax_sparse.BCOO.fromdense(B.T).astype(jnp.int32)
    C = jax_sparse.BCOO.fromdense(jnp.eye(nb_cons, nb_cons, dtype=jnp.int32))
    sparsity_pattern_right = jax_sparse.bcoo_concatenate(
        [BT_sparsity_pattern, C], dimension=0
    )

    sparsity_pattern_KKT = jax_sparse.bcoo_concatenate(
        [sparsity_pattern_left, sparsity_pattern_right], dimension=1
    )

    return sparsity_pattern_KKT



# Copyright (C) 2025 ETH Zurich (Mohit Pundir)
#
# This file is part of tatva.
#
# tatva is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tatva is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with tatva.  If not, see <https://www.gnu.org/licenses/>.


from __future__ import annotations

from enum import Enum
from typing import Literal, NamedTuple

import jax
import jax.numpy as jnp


class ElementType(Enum):
    """Enumeration of different finite element types."""

    TRIANGLE = "triangle"
    QUAD = "quad"
    TETRAHEDRON = "tetrahedron"
    HEXAHEDRON = "hexahedron"


class Mesh(NamedTuple):
    """A class used to represent a Mesh for finite element method (FEM) simulations.

    Attributes:
        nodes: The coordinates of the mesh nodes.
        elements: The connectivity of the mesh elements.
    """

    coords: jax.Array  # Shape (n_nodes, n_dim)
    elements: jax.Array  # Shape (n_elements, n_nodes_per_element)

    @classmethod
    def unit_square(
        cls,
        n_x: int,
        n_y: int,
        *,
        type: ElementType | Literal["triangle", "quad"] = ElementType.TRIANGLE,
        dim: Literal[2, 3] = 2,
    ) -> Mesh:
        """Generate a unit square mesh with n_x and n_y nodes in the x and y directions."""
        return cls.rectangle((0.0, 1.0), (0.0, 1.0), n_x, n_y, type=type, dim=dim)

    @classmethod
    def rectangle(
        cls,
        x: tuple[float, float],
        y: tuple[float, float],
        n_x: int,
        n_y: int,
        *,
        type: ElementType | Literal["triangle", "quad"] = ElementType.TRIANGLE,
        dim: Literal[2, 3] = 2,
    ) -> Mesh:
        """Generate a rectangular mesh with specified x and y ranges and number of nodes."""
        match ElementType(type):
            case ElementType.TRIANGLE:
                coords, elements = cls._rectangle_triangular(x, y, n_x, n_y, dim)
            case ElementType.QUAD:
                coords, elements = cls._rectangle_quadrilateral(x, y, n_x, n_y, dim)
            case _:
                raise NotImplementedError(f"Element type {type} not implemented.")

        return cls(coords=coords, elements=elements)

    @staticmethod
    def _rectangle_triangular(
        x: tuple[float, float], y: tuple[float, float], n_x: int, n_y: int, dim: int = 2
    ) -> tuple[jax.Array, jax.Array]:
        x_vals = jnp.linspace(x[0], x[1], n_x + 1)
        y_vals = jnp.linspace(y[0], y[1], n_y + 1)
        xv, yv = jnp.meshgrid(x_vals, y_vals, indexing="ij")
        coords = jnp.stack([xv.ravel(), yv.ravel()], axis=-1)

        def node_id(i, j):
            return i * (n_y + 1) + j

        elements = []
        for i in range(n_x):
            for j in range(n_y):
                n0 = node_id(i, j)
                n1 = node_id(i + 1, j)
                n2 = node_id(i, j + 1)
                n3 = node_id(i + 1, j + 1)
                elements.append([n0, n1, n3])
                elements.append([n0, n3, n2])

        if dim == 3:
            coords = jnp.hstack([coords, jnp.zeros((coords.shape[0], 1))])

        return coords, jnp.array(elements, dtype=jnp.int32)

    @staticmethod
    def _rectangle_quadrilateral(
        x: tuple[float, float], y: tuple[float, float], n_x: int, n_y: int, dim: int = 2
    ) -> tuple[jax.Array, jax.Array]:
        x_vals = jnp.linspace(x[0], x[1], n_x + 1)
        y_vals = jnp.linspace(y[0], y[1], n_y + 1)
        xv, yv = jnp.meshgrid(x_vals, y_vals, indexing="ij")
        coords = jnp.stack([xv.ravel(), yv.ravel()], axis=-1)

        def node_id(i, j):
            return i * (n_y + 1) + j

        elements = []
        for i in range(n_x):
            for j in range(n_y):
                n0 = node_id(i, j)
                n1 = node_id(i + 1, j)
                n2 = node_id(i + 1, j + 1)
                n3 = node_id(i, j + 1)
                elements.append([n0, n1, n2, n3])

        if dim == 3:
            coords = jnp.hstack([coords, jnp.zeros((coords.shape[0], 1))])

        return coords, jnp.array(elements, dtype=jnp.int32)


@jax.jit
def find_containing_polygons(
    points: jax.Array,
    polygons: jax.Array,
) -> jax.Array:
    """
    Finds the index of the containing polygon for each point.

    This function uses a vectorized Ray Casting algorithm and is JIT-compiled
    for maximum performance. It assumes polygons are non-overlapping.

    Args:
        points (Array): An array of points to test, shape (num_points, 2).
        polygons (Array): A 3D array of polygons, where each polygon is a list of
            vertices. Shape (num_polygons, num_vertices, 2).

    Returns:
        Array: An array of shape (num_points,) where each element is the index of the
            polygon containing the corresponding point. Returns -1 if a point is not in
            any polygon.
    """

    # --- Core function for a single point and a single polygon ---
    def is_inside(point, vertices):
        px, py = point

        # Get all edges of the polygon by pairing vertices with the next one
        p1s = vertices
        p2s = jnp.roll(vertices, -1, axis=0)  # Get p_{i+1} for each p_i

        # Conditions for a valid intersection of the horizontal ray from the point
        # 1. The point's y-coord must be between the edge's y-endpoints
        y_cond = (p1s[:, 1] <= py) & (p2s[:, 1] > py) | (p2s[:, 1] <= py) & (
            p1s[:, 1] > py
        )

        # 2. The point's x-coord must be to the left of the edge's x-intersection
        # Calculate the x-intersection of the ray with the edge
        x_intersect = (p2s[:, 0] - p1s[:, 0]) * (py - p1s[:, 1]) / (
            p2s[:, 1] - p1s[:, 1]
        ) + p1s[:, 0]
        x_cond = px < x_intersect

        # An intersection occurs if both conditions are met.
        intersections = jnp.sum(y_cond & x_cond)

        # The point is inside if the number of intersections is odd.
        return intersections % 2 == 1

    # --- Vectorize and apply the function ---
    # Create a boolean matrix: matrix[i, j] is True if point i is in polygon j
    # Vmap over points (axis 0) and polygons (axis 0)
    # in_axes=(0, None) -> maps over points, polygon is fixed
    # in_axes=(None, 0) -> maps over polygons, point is fixed
    # We vmap the second case over all points
    is_inside_matrix = jax.vmap(
        lambda p: jax.vmap(lambda poly: is_inside(p, poly))(polygons), in_axes=(0)
    )(points)

    # Find the index of the first 'True' value for each point (row).
    # This gives the index of the containing polygon.
    # We add a 'False' column to handle points outside all polygons.
    # jnp.argmax will then return the index of this last column.

    padded_matrix = jnp.pad(
        is_inside_matrix, ((0, 0), (0, 1)), "constant", constant_values=False
    )

    # If the point is not in any polygon, the all values in the row will be False.
    # We use this to map the index to -1.
    not_in_any_polygon = jnp.all(~is_inside_matrix, axis=1)

    indices = jnp.argmax(padded_matrix, axis=1)

    # we map the index to -1 if the point is not in any polygon.
    return jnp.where(not_in_any_polygon == True, -1, indices)



# Copyright (C) 2025 ETH Zurich (Mohit Pundir)
#
# This file is part of tatva.
#
# tatva is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tatva is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with tatva.  If not, see <https://www.gnu.org/licenses/>.


from abc import abstractmethod
from typing import TYPE_CHECKING

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array

if TYPE_CHECKING:
    from typing import ClassVar as AbstractClassVar
else:
    from equinox import AbstractClassVar


class Element(eqx.Module):
    """Base Module for all finite elements, compatible with JAX."""

    quad_points: AbstractClassVar[Array]
    quad_weights: AbstractClassVar[Array]

    @abstractmethod
    def shape_function(self, xi: Array) -> Array:
        """Returns the shape functions and their derivatives at a point."""
        raise NotImplementedError

    @abstractmethod
    def shape_function_derivative(self, xi: Array) -> Array:
        """Returns the derivative of the shape functions with respect to the local coordinates."""
        raise NotImplementedError

    def get_jacobian(self, xi: Array, nodal_coords: Array) -> tuple[Array, Array]:
        dNdr = self.shape_function_derivative(xi)
        J = dNdr @ nodal_coords
        return J, jnp.linalg.det(J)

    def interpolate(self, xi: Array, nodal_values: Array) -> Array:
        N = self.shape_function(xi)
        return N @ nodal_values

    def gradient(self, xi: Array, nodal_values: Array, nodal_coords: Array) -> Array:
        dNdr = self.shape_function_derivative(xi)
        J, _ = self.get_jacobian(xi, nodal_coords)
        dNdX = jnp.linalg.inv(J) @ dNdr
        return dNdX @ nodal_values

    def get_local_values(
        self, xi: Array, nodal_values: Array, nodal_coords: Array
    ) -> tuple[Array, Array, Array]:
        """Returns a tuple containing the interpolated value, gradient, and determinant of the Jacobian.

        Args:
            xi: Local coordinates (shape: (n_dim,)).
            nodal_values: Values at the nodes of the element (shape: (n_nodes, n_values)).
            nodal_coords: Coordinates of the nodes of the element (shape: (n_nodes, n_dim)).

        Returns:
            A tuple containing:
                - Interpolated value at the local coordinates (shape: (n_values,)).
                - Gradient of the nodal values at the local coordinates (shape: (n_dim, n_values)).
                - Determinant of the Jacobian (scalar).
        """
        N = self.shape_function(xi)
        dNdr = self.shape_function_derivative(xi)
        J, detJ = self.get_jacobian(xi, nodal_coords)
        dNdX = jnp.linalg.inv(J) @ dNdr
        return N @ nodal_values, dNdX @ nodal_values, detJ


class Line2(Element):
    """A 2-node linear interval element."""

    quad_points = jnp.array([[0.0]])
    quad_weights = jnp.array([2.0])

    def shape_function(self, xi: Array) -> Array:
        xi_val = xi[0]
        return jnp.array([0.5 * (1.0 - xi_val), 0.5 * (1.0 + xi_val)])

    def shape_function_derivative(self, xi: Array) -> Array:
        """Returns the derivative of the shape functions with respect to the local coordinates."""
        return jnp.array([-0.5, 0.5])

    def get_jacobian(self, xi: Array, nodal_coords: Array) -> tuple[Array, Array]:
        _N, dNdr = self.shape_function(xi), self.shape_function_derivative(xi)
        J = dNdr @ nodal_coords
        t = jnp.asarray([J[0], J[1]]) / jnp.linalg.norm(J)
        return jnp.dot(J, t), jnp.dot(J, t)

    def gradient(self, xi: Array, nodal_values: Array, nodal_coords: Array) -> Array:
        _N, dNdr = self.shape_function(xi), self.shape_function_derivative(xi)
        J, _ = self.get_jacobian(xi, nodal_coords)
        dNdX = dNdr / J
        return dNdX @ nodal_values

    def get_local_values(
        self, xi: Array, nodal_values: Array, nodal_coords: Array
    ) -> tuple[Array, Array, Array]:
        N, dNdr = self.shape_function(xi), self.shape_function_derivative(xi)
        J, detJ = self.get_jacobian(xi, nodal_coords)
        dNdX = dNdr / J
        return N @ nodal_values, dNdX @ nodal_values, detJ


class Tri3(Element):
    """A 3-node linear triangular element."""

    quad_points = jnp.array([[1.0 / 3, 1.0 / 3]])
    quad_weights = jnp.array([1.0 / 2])

    def shape_function(self, xi: Array) -> Array:
        """Returns the shape functions evaluated at the local coordinates (xi, eta)."""
        xi1, xi2 = xi
        return jnp.array([1.0 - xi1 - xi2, xi1, xi2])

    def shape_function_derivative(self, *_args, **_kwargs) -> Array:
        """Returns the derivative of the shape functions with respect to the local coordinates (xi, eta)."""
        # shape (n_q, 2, 3)
        return jnp.array([[-1.0, -1.0], [1.0, 0.0], [0.0, 1.0]]).T


def _get_quad4_quadrature() -> tuple[Array, Array]:
    xi_vals = jnp.array([-1.0 / jnp.sqrt(3), 1.0 / jnp.sqrt(3)])
    w_vals = jnp.array([1.0, 1.0])
    quad_points = jnp.stack(jnp.meshgrid(xi_vals, xi_vals), axis=-1).reshape(-1, 2)
    weights = jnp.kron(w_vals, w_vals)
    return quad_points, weights


_quad4_qp, _quad4_w = _get_quad4_quadrature()


class Quad4(Element):
    """A 4-node bilinear quadrilateral element."""

    quad_points = _quad4_qp
    quad_weights = _quad4_w

    def shape_function(self, xi: Array) -> Array:
        r, s = xi
        return 0.25 * jnp.array(
            [(1 - r) * (1 - s), (1 + r) * (1 - s), (1 + r) * (1 + s), (1 - r) * (1 + s)]
        )

    def shape_function_derivative(self, xi: Array) -> Array:
        """Returns the derivative of the shape functions with respect to the local coordinates (xi, eta)."""
        r, s = xi
        return (
            0.25
            * jnp.array(
                [
                    [-(1 - s), -(1 - r)],
                    [(1 - s), -(1 + r)],
                    [(1 + s), (1 + r)],
                    [-(1 + s), (1 - r)],
                ]
            ).T
        )


class Tetrahedron4(Element):
    """A 4-node linear tetrahedral element."""

    quad_points = jnp.array([[1.0 / 4, 1.0 / 4, 1.0 / 4]])
    quad_weights = jnp.array([1.0 / 6])

    def shape_function(self, xi: Array) -> Array:
        """Returns the shape functions evaluated at the local coordinates (xi, eta, zeta)."""
        xi, eta, zeta = xi
        return jnp.array([1.0 - xi - eta - zeta, xi, eta, zeta])

    def shape_function_derivative(self, *_args, **_kwargs) -> Array:
        """Returns the derivative of the shape functions with respect to the local coordinates (xi, eta, zeta)."""
        # shape (n_q, 3, 4)
        return jnp.array(
            [[-1.0, -1.0, -1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]
        ).T


def _get_hex8_quadrature() -> tuple[Array, Array]:
    xi_vals = jnp.array([-1.0 / jnp.sqrt(3), 1.0 / jnp.sqrt(3)])
    quad_points = jnp.stack(jnp.meshgrid(xi_vals, xi_vals, xi_vals), axis=-1).reshape(
        -1, 3
    )
    weights = jnp.full(quad_points.shape[0], fill_value=1.0)
    return quad_points, weights


_hex8_qp, _hex8_w = _get_hex8_quadrature()


class Hexahedron8(Element):
    """A 8-node linear hexahedral element."""

    quad_points = _hex8_qp
    quad_weights = _hex8_w

    def shape_function(self, xi: Array) -> Array:
        """Returns the shape functions evaluated at the local coordinates (xi, eta, zeta)."""
        xi, eta, zeta = xi
        return (1 / 8) * jnp.array(
            [
                (1 - xi) * (1 - eta) * (1 - zeta),
                (1 + xi) * (1 - eta) * (1 - zeta),
                (1 + xi) * (1 + eta) * (1 - zeta),
                (1 - xi) * (1 + eta) * (1 - zeta),
                (1 - xi) * (1 - eta) * (1 + zeta),
                (1 + xi) * (1 - eta) * (1 + zeta),
                (1 + xi) * (1 + eta) * (1 + zeta),
                (1 - xi) * (1 + eta) * (1 + zeta),
            ]
        )

    def shape_function_derivative(self, xi: Array) -> Array:
        """Returns the derivative of the shape functions with respect to the local coordinates (xi, eta, zeta)."""
        # shape (n_q, 3, 8)
        xi, eta, zeta = xi
        return (1 / 8) * jnp.array(
            [
                [
                    -(1 - eta) * (1 - zeta),
                    -(1 - xi) * (1 - zeta),
                    -(1 - xi) * (1 - eta),
                ],
                [(1 - eta) * (1 - zeta), -(1 + xi) * (1 - zeta), -(1 + xi) * (1 - eta)],
                [(1 + eta) * (1 - zeta), (1 + xi) * (1 - zeta), -(1 + xi) * (1 + eta)],
                [-(1 + eta) * (1 - zeta), (1 - xi) * (1 - zeta), -(1 - xi) * (1 + eta)],
                [-(1 - eta) * (1 + zeta), -(1 - xi) * (1 + zeta), (1 - xi) * (1 - eta)],
                [(1 - eta) * (1 + zeta), -(1 + xi) * (1 + zeta), (1 + xi) * (1 - eta)],
                [(1 + eta) * (1 + zeta), (1 + xi) * (1 + zeta), (1 + xi) * (1 + eta)],
                [-(1 + eta) * (1 + zeta), (1 - xi) * (1 + zeta), (1 - xi) * (1 + eta)],
            ]
        )




# Copyright (C) 2025 ETH Zurich (Mohit Pundir)
#
# This file is part of tatva.
#
# tatva is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tatva is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with tatva.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np
import pyvista as pv


def get_pyvista_grid(mesh, cell_type="quad"):
    if mesh.coords.shape[1] == 2:
        pv_points = np.hstack((mesh.coords, np.zeros(shape=(mesh.coords.shape[0], 1))))
    else:
        pv_points = mesh.coords

    cell_type_dict = {
        "quad": 4,
        "triangle": 3,
        "tetra": 4,
    }

    pv_cells = np.hstack(
        (
            np.full(
                fill_value=cell_type_dict[cell_type], shape=(mesh.elements.shape[0], 1)
            ),
            mesh.elements,
        )
    )

    pv_cell_type_dict = {
        "quad": pv.CellType.QUAD,
        "triangle": pv.CellType.TRIANGLE,
        "tetra": pv.CellType.TETRA,
    }
    cell_types = np.full(
        fill_value=pv_cell_type_dict[cell_type], shape=(mesh.elements.shape[0],)
    )

    grid = pv.UnstructuredGrid(pv_cells, cell_types, pv_points)

    return grid



import jax
import jax.numpy as jnp
import numpy as np
import pytest

from tatva.element import Tri3
from tatva.mesh import Mesh
from tatva.operator import Operator

jax.config.update("jax_enable_x64", True)

NODES = jnp.array(
    [
        [0.0, 0.0],
        [1.0, 0.0],
        [1.0, 1.0],
        [0.0, 1.0],
    ],
    dtype=jnp.float64,
)

ELEMENTS = jnp.array(
    [
        [0, 1, 2],
        [0, 2, 3],
    ],
    dtype=jnp.int32,
)

EXPECTED_ELEMENT_AREAS = np.array([0.5, 0.5], dtype=np.float64)


def test_operator_raises_for_dimension_mismatch():
    coords = jnp.array(
        [
            [0.0, 0.0, 0.0],
            [1.0, 0.0, 0.0],
            [0.0, 1.0, 0.0],
        ],
        dtype=jnp.float64,
    )
    elements = jnp.array([[0, 1, 2]], dtype=jnp.int32)
    mesh = Mesh(coords=coords, elements=elements)

    with pytest.raises(ValueError, match="expects 2D coordinates"):
        Operator(mesh, Tri3())


def test_operator_raises_for_node_count_mismatch():
    coords = jnp.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [1.0, 1.0],
            [0.0, 1.0],
        ],
        dtype=jnp.float64,
    )
    elements = jnp.array([[0, 1, 2, 3]], dtype=jnp.int32)
    mesh = Mesh(coords=coords, elements=elements)

    with pytest.raises(ValueError, match="lists 4 nodes per element"):
        Operator(mesh, Tri3())


def test_operator_raises_for_invalid_connectivity():
    coords = jnp.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [0.0, 1.0],
        ],
        dtype=jnp.float64,
    )
    elements = jnp.array([[0, 1, 3]], dtype=jnp.int32)
    mesh = Mesh(coords=coords, elements=elements)

    with pytest.raises(ValueError, match="outside the mesh coordinates"):
        Operator(mesh, Tri3())


@pytest.fixture(scope="module")
def tri_operator() -> Operator:
    mesh = Mesh(coords=NODES, elements=ELEMENTS)
    return Operator(mesh, Tri3())


def _manual_map_results(func, nodal_values: jnp.ndarray) -> np.ndarray:
    results = []
    for element in np.array(ELEMENTS):
        element_values = np.array(nodal_values)[element]
        element_results = [
            np.array(func(xi, element_values)) for xi in np.array(Tri3().quad_points)
        ]
        results.append(np.asarray(element_results))
    return np.asarray(results)


def test_map_applies_function_per_element_and_quad(tri_operator: Operator):
    """Test that the map method applies the function to each element and quadrature point
    correctly."""
    nodal_values = jnp.arange(NODES.shape[0], dtype=jnp.float64)

    def func(xi, el_values):
        return jnp.sum(el_values) + jnp.sum(xi)

    mapped = tri_operator.map(func)
    result = mapped(nodal_values)

    expected = _manual_map_results(
        lambda xi, el_vals: np.sum(el_vals) + np.sum(np.array(xi)),
        nodal_values,
    )
    np.testing.assert_allclose(result, expected)


def test_map_respects_element_quantity(tri_operator: Operator):
    """Test that the element_quantity argument correctly passes additional per-element
    data to the mapped function."""
    nodal_values = jnp.arange(NODES.shape[0], dtype=jnp.float64)
    element_bias = jnp.array([10.0, 20.0], dtype=jnp.float64)

    def func(xi, el_values, bias):
        return jnp.sum(el_values) + bias + xi[0]

    mapped = tri_operator.map(func, element_quantity=(1,))
    result = mapped(nodal_values, element_bias)

    def manual(xi, el_vals, bias):
        return np.sum(el_vals) + bias + np.array(xi)[0]

    expected = []
    for idx, element in enumerate(np.array(ELEMENTS)):
        element_values = np.array(nodal_values)[element]
        element_results = [
            manual(xi, element_values, float(element_bias[idx]))
            for xi in np.array(Tri3().quad_points)
        ]
        expected.append(np.asarray(element_results))

    np.testing.assert_allclose(result, np.asarray(expected))


def test_map_over_elements_matches_manual(tri_operator: Operator):
    """Test that map_over_elements applies the function to each element correctly."""
    nodal_values = jnp.arange(NODES.shape[0], dtype=jnp.float64)

    def func(el_values):
        return jnp.sum(el_values)

    mapped = tri_operator.map_over_elements(func)
    result = mapped(nodal_values)

    expected = np.array(
        [np.sum(np.array(nodal_values)[element]) for element in np.array(ELEMENTS)]
    )
    np.testing.assert_allclose(result, expected)


def test_eval_returns_weighted_average(tri_operator: Operator):
    nodal_values = jnp.array([0.0, 1.0, 2.0, 3.0], dtype=jnp.float64)
    result = tri_operator.eval(nodal_values)
    expected = np.array([[1.0], [5.0 / 3.0]], dtype=np.float64)
    np.testing.assert_allclose(result, expected)


def test_grad_of_linear_field_is_constant(tri_operator: Operator):
    coeffs = jnp.array([2.0, 3.0])
    nodal_values = NODES @ coeffs
    grad = tri_operator.grad(nodal_values)
    expected = np.array([[[2.0, 3.0]], [[2.0, 3.0]]], dtype=np.float64)
    np.testing.assert_allclose(grad, expected)


def test_integrate_nodal_field_matches_area(tri_operator: Operator):
    nodal_values = jnp.ones(NODES.shape[0], dtype=jnp.float64)
    per_element = tri_operator.integrate_per_element(nodal_values)
    total = tri_operator.integrate(nodal_values)

    np.testing.assert_allclose(per_element, EXPECTED_ELEMENT_AREAS)
    np.testing.assert_allclose(total, np.sum(EXPECTED_ELEMENT_AREAS))


def test_integrate_quad_values_matches_manual(tri_operator: Operator):
    quad_values = jnp.full(
        (ELEMENTS.shape[0], Tri3().quad_points.shape[0]), 4.0, dtype=jnp.float64
    )
    per_element = tri_operator.integrate_per_element(quad_values)
    expected = EXPECTED_ELEMENT_AREAS * 4.0
    np.testing.assert_allclose(per_element, expected)


def test_interpolate_recovers_linear_function(tri_operator: Operator):
    nodal_values = jnp.sum(NODES, axis=1)
    points = jnp.array(
        [
            [0.25, 0.25],
            [0.75, 0.25],
            [0.25, 0.75],
            [0.5, 0.5],
        ],
        dtype=jnp.float64,
    )
    interpolated = tri_operator.interpolate(nodal_values, points)
    expected = np.sum(np.array(points), axis=1)
    np.testing.assert_allclose(interpolated, expected)


def test_interpolate_raises_for_points_outside_mesh(tri_operator: Operator):
    nodal_values = jnp.ones(NODES.shape[0], dtype=jnp.float64)
    points = jnp.array([[1.5, 0.5]], dtype=jnp.float64)
    with pytest.raises(RuntimeError):
        tri_operator.interpolate(nodal_values, points)



# Copyright (C) 2025 ETH Zurich (Mohit Pundir)
#
# This file is part of tatva.
#
# tatva is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# tatva is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with tatva.  If not, see <https://www.gnu.org/licenses/>.


from __future__ import annotations

from typing import Any, Callable, Generator, Self, overload

import jax.numpy as jnp
from jax import Array
from jax.tree_util import register_pytree_node_class


class _field:
    """A descriptor to define fields in the State class."""

    shape: tuple[int, ...]
    default_factory: Callable | None

    def __init__(
        self,
        shape: tuple[int, ...],
        default_factory: Callable | None = None,
        slice: slice | None = None,
    ) -> None:
        self.shape = shape if len(shape) > 1 else (*shape, 1)
        self.default_factory = default_factory
        self.slice: slice = slice  # type: ignore

    def __getitem__(self, arg) -> Array:
        # Normalize to tuple
        if not isinstance(arg, tuple):
            arg = (arg,)
        # extend with full slices
        if len(arg) < len(self.shape):
            arg = arg + (slice(None),) * (len(self.shape) - len(arg))
        # build index arrays
        idxs = []
        for dim, sub in enumerate(arg):
            if isinstance(sub, slice):
                start, stop, step = sub.indices(self.shape[dim])
                idxs.append(jnp.arange(start, stop, step))
            elif isinstance(sub, (int, jnp.integer)):
                idxs.append(jnp.array([sub]))
            else:
                idxs.append(jnp.asarray(sub))

        mesh = jnp.meshgrid(*idxs, indexing="ij")
        multi_idx = [m.flatten() for m in mesh]
        flat_local = jnp.ravel_multi_index(multi_idx, dims=self.shape)

        return jnp.array(flat_local + self.slice.start)


class field(_field):
    @overload
    def __get__(self, instance: None, owner) -> field: ...
    @overload
    def __get__(self, instance: Compound, owner) -> Array: ...
    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        # get slice
        return instance.arr[self.slice].reshape(self.shape)

    def __set__(self, instance: Compound, value: Array | float | int) -> None:
        arr = jnp.asarray(value)
        instance.arr = instance.arr.at[self.slice].set(arr.flatten())

    def __delete__(self, instance):
        raise AttributeError(
            f"Cannot delete field ... from {instance.__class__.__name__}."
        )


class _field_sub_of_stack(field):
    """A descriptor to define fields that are sub-fields of a stacked field in the State class."""

    def __init__(
        self,
        shape: tuple[int, ...],
        parent_field: _field,
        parent_slice: tuple[slice, ...],
    ) -> None:
        super().__init__(shape)
        self.parent_field = parent_field
        self.parent_slice = parent_slice

    @overload
    def __get__(self, instance: None, owner) -> field: ...
    @overload
    def __get__(self, instance: Compound, owner) -> Array: ...
    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        # get slice
        return instance.arr[self.parent_field.slice].reshape(self.parent_field.shape)[
            self.parent_slice
        ]

    def __set__(self, instance: Compound, value: Array | float | int) -> None:
        arr = jnp.asarray(value)
        instance.arr = (
            instance.arr[self.parent_field.slice]
            .reshape(self.parent_field.shape)
            .at[self.parent_slice]
            .set(arr)
        ).flatten()

    def __getitem__(self, arg) -> Array:
        # Normalize to tuple
        if not isinstance(arg, tuple):
            arg = (arg,)
        # extend with full slices
        if len(arg) < len(self.shape):
            arg = arg + (slice(None),) * (len(self.shape) - len(arg))

        # get indices in the parent field
        parent_idxs = self.parent_field.__getitem__(slice(None)).reshape(
            self.parent_field.shape
        )
        return parent_idxs[self.parent_slice].__getitem__(arg).flatten()


class _CompoundMeta(type):
    fields: tuple[tuple[str, field], ...]
    size: int

    def __new__(
        mcls, name: str, bases: tuple[type, ...], namespace: dict[str, Any], **kwargs
    ):
        fields: list[tuple[str, field]] = []
        size: int = 0

        # find all fields in the namespace and compute their slices in the flat array
        for attr_name, attr_value in namespace.items():
            if isinstance(attr_value, field):
                n = int(jnp.prod(jnp.asarray(attr_value.shape)))
                # get slice indices
                start = size
                end = start + n
                attr_value.slice = slice(start, end)

                fields.append((attr_name, attr_value))
                size += n

        cls = super().__new__(mcls, name, bases, namespace)
        cls.fields = tuple(fields)
        cls.size = size
        if kwargs.get("stack_fields") is not None:
            cls._stack_fields(**kwargs)  # type: ignore

        # register as pytree node for JAX transformations
        register_pytree_node_class(cls)
        return cls

    def __getitem__(cls, arg) -> Array:
        node, *dofs = arg if isinstance(arg, tuple) else (arg,)
        nodal_vals = jnp.hstack(
            [f[node].reshape(-1, *f.shape[1:]) for _, f in cls.fields],
            dtype=int,
        )
        return (nodal_vals[:, *dofs] if dofs else nodal_vals).flatten()


class Compound(metaclass=_CompoundMeta):
    """A compound array/state.

    A helper class to create a compound state with multiple fields. It simplifies packing
    and unpacking into and from a flat array. Useful to manage fields while working with a
    flat array for the solver.

    Args:
        arr: The flat data array. If None, initializes to zeros.

    Examples:

    Create a compound state with fields::

        class MyState(Compound):
            u = field(shape=(N, 3))
            phi = field(shape=(N,), default_factory=lambda: jnp.ones(N))

        state = MyState()

    Use `state.arr` to access the flat array, and `state.u`, `state.phi` to access the
    individual fields.

    You can use iterator unpacking to directly unpack the fields from the state::

        u, phi = MyState(arr)

    """

    arr: Array
    size: int = 0

    def tree_flatten(self) -> tuple[tuple[Array], Any]:
        return (self.arr,), None

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: tuple[Array]) -> Self:
        return cls(*children)

    @classmethod
    def _stack_fields(cls, stack_fields: tuple[str, ...], stack_axis: int = -1) -> None:
        """Reorder fields by stacking specified fields along some axis. Modifies the class
        in place!

        Args:
            stack_fields: Names of the fields to be stacked.
            stack_axis: Axis along which to stack the fields. Defaults to -1.
        """
        fields = {k: v for k, v in cls.fields}
        base_shape = fields[stack_fields[0]].shape
        stack_axis = stack_axis % len(base_shape)  # get positive axis
        base_shape = jnp.asarray(base_shape)[
            jnp.array([i for i in range(len(base_shape)) if i != stack_axis])
        ]

        stack_size = sum(
            int(jnp.prod(jnp.asarray(fields[name].shape))) // base_shape.prod()
            for name in stack_fields
        )
        stacked_shape = jnp.insert(base_shape, stack_axis, stack_size)
        size = int(jnp.prod(jnp.asarray(stacked_shape)))
        stacked_slice = slice(0, size)
        stacked_field = _field(tuple(stacked_shape), slice=stacked_slice)

        new_fields = []

        # create new fields that are sub-fields of the stacked field
        for idx, name in enumerate(stack_fields):
            new_field = _field_sub_of_stack(
                shape=fields[name].shape,
                parent_field=stacked_field,
                parent_slice=tuple(
                    slice(None)
                    if i != stack_axis
                    else slice(
                        sum(
                            int(jnp.prod(jnp.asarray(fields[n].shape)))
                            // base_shape.prod()
                            for n in stack_fields[:idx]
                        ),
                        sum(
                            int(jnp.prod(jnp.asarray(fields[n].shape)))
                            // base_shape.prod()
                            for n in stack_fields[: idx + 1]
                        ),
                    )
                    for i in range(len(stacked_shape))
                ),
            )
            # check base shape compatibility or raise error
            if not jnp.all(
                jnp.asarray(new_field.shape)[
                    jnp.array(
                        [i for i in range(len(new_field.shape)) if i != stack_axis]
                    )
                ]
                == base_shape
            ):
                raise ValueError(
                    f"Field {name} with shape {new_field.shape} is not compatible with "
                    f"base shape {base_shape} along axis {stack_axis}."
                )

            # set new field in class and add to new fields list
            setattr(cls, name, new_field)
            new_fields.append((name, new_field))

        # update other fields slices
        offset = size
        for name, field_obj in cls.fields:
            if name in stack_fields:
                continue
            n = int(jnp.prod(jnp.asarray(field_obj.shape)))
            start = offset
            end = start + n
            field_obj.slice = slice(start, end)
            offset += n
            new_fields.append((name, field_obj))

        # finally update class fields and size
        cls.fields = tuple(new_fields)

    def __init__(self, arr: Array | None = None) -> None:
        """Initialize the state with given keyword arguments."""
        if arr is not None:
            assert arr.size == self.size, (
                f"Array size {arr.size} does not match expected size {self.size}."
            )
            self.arr = arr
        else:
            self.arr = jnp.zeros(self.size, dtype=float)

    def __len__(self) -> int:
        return len(self.fields)

    def __iter__(self) -> Generator[Array, None, None]:
        for name, _ in self.fields:
            yield getattr(self, name)

    def __repr__(self) -> str:
        # print shape of each field in the class
        field_reprs = [
            f"{name}={getattr(type(self), name).shape}" for name, _ in self.fields
        ]
        return f"{self.__class__.__name__}({', '.join(field_reprs)})"

    def __add__(self, other: Self) -> Self:
        return self.__class__(self.arr + other.arr)
